#ifndef __GRAID_H_INCLUDED__
#define __GRAID_H_INCLUDED__

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <linux/iommu.h>
#include <linux/mm.h>
#include <linux/blkdev.h>
#include <linux/blk-mq.h>
#include <linux/nvme.h>
#include <linux/dmapool.h>
#include <linux/completion.h>
#include <linux/workqueue.h>
#include <linux/mempool.h>
#include <linux/io-64-nonatomic-lo-hi.h>
#include <linux/async.h>

#include "ioctl.h"
#include "dmapool.h"

/* module.c */
extern struct device* graid_device;
extern int marked_clean_shutdown;

/* virtual-nvme.c */
extern int graid_init_dg_queue(void);
extern void graid_cleanup_dg_queue(void);
extern int graid_vir_nvme_init(void);
extern void graid_vir_nvme_exit(void);
extern void graid_vir_nvme_cleanup(void);
extern bool graid_vir_nvme_ready(void);
extern int graid_create_dg(struct graid_nvme_drive_group __user *ndg);
extern int graid_destroy_dg(int dg_id);
extern int graid_create_ctrl(struct graid_virt_nvme_dev __user *vdev);
extern int graid_destroy_ctrl(int dg_vd_id);
extern int graid_start_queues(u32 dg_vd_id, u32 optimal);
extern int graid_stop_queues(int dg_vd_id);
extern int graid_check_ctrl(int dg_vd_id);
extern int graid_online_cores(void);
extern int graid_dgq_node(int);

/* user.c */
extern struct file_operations graid_fops;

#define GRAID_DMA_NODE_OFFSET 10
/* dma.c */
extern        int graid_is_dma_pool(uint64_t addr, uint64_t len);
extern dma_addr_t graid_pci_get_cq_paddr(int instance, int qid);
extern dma_addr_t graid_pci_get_sq_paddr(int instance, int qid);
extern dma_addr_t graid_vir_get_cq_paddr(int instance, int qid);
extern dma_addr_t graid_vir_get_sq_paddr(int instance, int qid);
extern      void* graid_pci_get_cq_vaddr(int instance, int qid);
extern      void* graid_pci_get_sq_vaddr(int instance, int qid);
extern      void* graid_vir_get_cq_vaddr(int instance, int qid);
extern      void* graid_vir_get_sq_vaddr(int instance, int qid);
extern      void* graid_vir_get_dbs_vaddr(int dg_id);
extern      void* graid_blk_get_dbs_vaddr(int pd_id);
extern        int init_dma_pools(struct device *dev);
extern       void cleanup_dma_pools(struct device *dev);
extern	     long graid_ioctl_get_dma_pool(void __user* pool);

extern int graid_alloc_pages(int node, u32 page_nr, void **page_kptr_list);
extern void graid_free_page(int node, void *page_kptr);
extern void graid_free_pages(int node, u32 page_nr, void **page_kptr_list);
extern dma_addr_t graid_map_page(struct page *page);
extern int graid_map_dmapool_pages(struct vm_area_struct *vma, int node, size_t start);

extern int graid_ioctl_get_mem_addrs(void __user *arg);
extern int graid_ioctl_dma_set_bar1(void __user *arg);
extern int graid_ioctl_dma_reset_bar1(void __user *arg);
extern void graid_unmap_bar1(void);
extern void graid_free_gpu_page(int node, void *page_kptr, u64 page_pptr);
extern int graid_alloc_gpu_page(int node, void **page_kptr, uint64_t *page_pptr);

/* pci-nvme.c */
extern int graid_pci_nvme_init(void);
extern void graid_pci_nvme_exit(void);
extern int graid_is_pci_nvme_bar0(uint64_t addr, uint64_t len);
extern int graid_pci_nvme_query(struct graid_pci_nvme_dev __user *arg);
extern int graid_pci_nvme_reset(unsigned long instance);
extern int graid_pci_alloc_id(void __user *arg);
extern int graid_pci_free_id(unsigned long instance);
extern int __graid_pci_alloc_id(void);
extern void __graid_pci_free_id(int id);
extern int graid_pci_nvme_dealloc(unsigned long instance);

/* blk-forward.c */
extern int graid_blk_init(void);
extern void graid_blk_exit(void);
extern void graid_blk_fwd_cleanup(void);
extern int graid_blk_create(void __user *arg);
extern int graid_blk_destroy(int pd_id);
extern int graid_blk_reset(int pd_id);

/* nvidia.c */
int graid_nvidia_init(void);
void graid_nvidia_exit(void);
int graid_gpu_pin_bar1(struct graid_gpu_bar1_req __user* user_req);
int graid_gpu_unpin_bar1(struct graid_gpu_bar1_req __user* user_req);
void graid_unpin_and_release(void);
void* graid_bar1_p2k(uint64_t paddr);

#endif
