// -*- mode:c; tab-width:8; c-basic-offset:8; indent-tabs-mode:t -*-
// vim: ts=8 sw=8 smarttab
#include "mpmc_queue.h"
#include <linux/vmalloc.h>

mpmc_queue_t *mpmc_queue_alloc(size_t items)
{
	size_t alloc_len = sizeof(mpmc_queue_t) + (sizeof(mpmc_queue_cell_t) * items);
	mpmc_queue_t *queue = NULL;
	size_t i;

	if ((items < 2) || ((items & (items - 1)) != 0)) {
		pr_err("%s: item number must be power of 2.\n", __func__);
		return NULL;
	}

	queue = vmalloc(alloc_len);
	if (unlikely(queue == NULL))
		return NULL;

	queue->cell_mask = items - 1;
	atomic64_set(&queue->enqueue_pos, 0);
	atomic64_set(&queue->dequeue_pos, 0);
	for (i = 0; i < items; ++i)
		atomic64_set(&queue->cells[i].seq, i);

	return queue;
}

void mpmc_queue_free(mpmc_queue_t *queue)
{
	vfree(queue);
}

bool mpmc_enqueue(mpmc_queue_t *queue, uint64_t val)
{
    mpmc_queue_cell_t* cell;
    size_t pos = atomic64_read(&queue->enqueue_pos);
    for (;;)
    {
      uint64_t seq;
      int64_t dif;
      cell = &queue->cells[pos & queue->cell_mask];
      seq = atomic64_read_acquire(&cell->seq);
      dif = (int64_t)seq - (int64_t)pos;
      if (dif == 0)
      {
        if (atomic64_cmpxchg(&queue->enqueue_pos, pos, pos + 1) == pos)
          break;
      }
      else if (dif < 0)
        return false;
      else
        pos = atomic64_read(&queue->enqueue_pos);
    }
    cell->val = val;
    atomic64_set_release(&cell->seq, pos + 1);
    return true;
}

bool mpmc_dequeue(mpmc_queue_t *queue, uint64_t *val)
{
	mpmc_queue_cell_t* cell;
	size_t pos = atomic64_read(&queue->dequeue_pos);
	for (;;)
	{
		size_t seq;
		int64_t dif;
		cell = &queue->cells[pos & queue->cell_mask];
		seq = atomic64_read_acquire(&cell->seq);
		dif = (int64_t)seq - (int64_t)(pos + 1);
		if (dif == 0)
		{
			if (atomic64_cmpxchg(&queue->dequeue_pos, pos, pos + 1) == pos)
				break;
		}
		else if (dif < 0)
			return false;
		else
			pos = atomic64_read(&queue->dequeue_pos);
	}
	*val = cell->val;
        atomic64_set_release(&cell->seq, pos + queue->cell_mask + 1);
	return true;
}


